use crate::model::{Entity, EntityColumns, EntityID};
use crate::DB;

pub mod condition;

/// Wraps an entity with its ID, for example as a query result.
///
/// The wrapped value is accessible via `Deref`, so this should be mostly
/// transparent.
#[derive(Debug)]
pub struct WithID<T: Entity> {
    wrap: T,
    id: <T as Entity>::ID,
}

impl<T: Entity> WithID<T> {
    fn wrap(what: T, raw_id: i64) -> Self {
        Self {
            wrap: what,
            id: <T as Entity>::ID::from_raw_id(raw_id),
        }
    }
}

impl<T: Entity> WithID<T> {
    pub fn id(&self) -> <T as Entity>::ID {
        self.id
    }
}

impl<T: Entity> AsRef<T> for WithID<T> {
    fn as_ref(&self) -> &T {
        &self.wrap
    }
}

impl<T: Entity> std::ops::Deref for WithID<T> {
    type Target = T;
    fn deref(&self) -> &Self::Target {
        &self.wrap
    }
}

impl<T: Entity> std::ops::DerefMut for WithID<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.wrap
    }
}

/// Search for an entity by a property
pub fn get_one_by<T: Entity<Column = C>, C: EntityColumns<Entity = T>, V: rusqlite::ToSql>(
    db: &DB,
    c: C,
    val: V,
) -> Option<WithID<T>> {
    let table_name = <T as Entity>::table_name();
    let column_name = <T as Entity>::name(c);

    let mut prepared = db
        .conn
        .prepare(&format!(
            "SELECT * FROM \"{}\" WHERE \"{}\" = ?1",
            table_name, column_name
        ))
        .ok()?;

    let result = prepared.query_row([&val], |row| {
        let mut deser = crate::model::load::RowDeserializer::from_row(row);
        Ok(WithID::wrap(
            T::deserialize(&mut deser).expect("deserialization works"),
            row.get(0).expect("can get id"),
        ))
    });

    result.ok()
}

/// Search for all entities matching a property
pub fn get_all_by<T: Entity<Column = C>, C: EntityColumns<Entity = T>, V: rusqlite::ToSql>(
    db: &DB,
    c: C,
    val: V,
) -> Option<Vec<WithID<T>>> {
    let table_name = <T as Entity>::table_name();
    let column_name = <T as Entity>::name(c);

    let mut prepared = db
        .conn
        .prepare(&format!(
            "SELECT * FROM \"{}\" WHERE \"{}\" = ?1",
            table_name, column_name
        ))
        .ok()?;

    let rows = prepared
        .query_map([&val], |row| {
            let mut deser = crate::model::load::RowDeserializer::from_row(row);
            Ok(WithID::wrap(
                T::deserialize(&mut deser)?,
                row.get(0).expect("can get rowid"),
            ))
        })
        .ok()?;

    Some(rows.map(|x| x.unwrap()).collect())
}

/// Search for an entity by ID
pub fn get_one_by_id<T: Entity>(db: &DB, id: <T as Entity>::ID) -> Option<WithID<T>> {
    let table_name = <T as Entity>::table_name();
    let mut prepared = db
        .conn
        .prepare(&format!("SELECT * FROM \"{}\" WHERE id = ?1", table_name))
        .ok()?;

    let result = prepared.query_row([&id], |row| {
        let mut deser = crate::model::load::RowDeserializer::from_row(row);
        Ok(WithID::wrap(
            T::deserialize(&mut deser).expect("deserialization works"),
            row.get(0).expect("can get rowid"),
        ))
    });

    result.ok()
}

/// Add an entity to its table
pub fn add<T: Entity + serde::Serialize>(db: &DB, m: &T) -> Option<<T as Entity>::ID> {
    let row = crate::model::store::serialize_as_row(m);

    let placeholders = (0..(<T as Entity>::column_count() - 1))
        .map(|n| format!("?{}", n + 1))
        .collect::<Vec<_>>()
        .join(",");

    let res = db.conn.prepare(&format!(
        "INSERT INTO \"{}\" VALUES (NULL, {})",
        <T as Entity>::table_name(),
        placeholders
    ));
    let mut prepared = res.ok()?;

    // make sure we bound enough things (not including ID column here)
    assert_eq!(row.len(), <T as Entity>::column_count() - 1);

    let id = prepared.insert(rusqlite::params_from_iter(row)).ok()?;
    Some(<T as Entity>::ID::from_raw_id(id))
}
